import logging

import numpy as np


class KmerLookup:
    def __init__(self, kmers, representative_kmers, lookup):
        self._kmers = kmers
        self._representative_kmers = representative_kmers
        self._lookup = lookup

    def max_node_id(self):
        return np.max(self._lookup)

    def to_file(self, file_name):
        np.savez(file_name, 
                 kmers = self._kmers,
                 representative_kmers = self._representative_kmers,
                 lookup = self._lookup)

    @classmethod
    def from_file(cls, file_name):
        data = np.load(file_name)
        return cls(data["kmers"],
                   data["representative_kmers"],
                   data["lookup"])

    def index_kmers(self):
        self._kmers.sort()
        self._representative_kmers = np.searchsorted(self._kmers, self._representative_kmers)

    def _get_indexes(self, kmers):
        return np.searchsorted(self._kmers, kmers)

    def count_kmers(self, kmers):
        indexes = self._get_indexes(kmers)
        return np.bincount(indexes[kmers==self._kmers[np.minimum(indexes, self._kmers.size-1)]],
                           minlength=self._kmers.size)
        
    def get_node_counts(self, kmers):
        counts = self.count_kmers(kmers)
        cum_counts = np.insert(np.cumsum(counts[self._representative_kmers]), 0, 0)
        node_counts = cum_counts[self._lookup[:, 1]]-cum_counts[self._lookup[:, 0]]
        return node_counts

class SimpleKmerLookup(KmerLookup):
    def get_node_counts(self, kmers):
        counts = self.count_kmers(kmers)
        return np.bincount(self._lookup, counts[self._representative_kmers])

    @classmethod
    def from_old_index_files(cls, filename):
        logging.info("From old index files")
        data = np.load(filename)
        kmers = data["kmers"]
        unique_kmers = np.unique(kmers)
        k = cls(unique_kmers, kmers, data["nodes"])
        k.index_kmers()
        return k

class AdvancedKmerLookup(SimpleKmerLookup):
    n_bins = 200000000
    max_value = 4**31
    def _get_indexes(self, kmers):
        return self._indexed_lookup.find_queries(kmers)

    def index_kmers(self):
        self._kmers.sort()
        self._representative_kmers = np.searchsorted(self._kmers, self._representative_kmers)
        self._indexed_lookup = IndexedSortedLookup.from_values_and_mod(
            self._kmers, self.max_value//self.n_bins+1,
            self.max_value, do_sort=False)

class Advanced2(AdvancedKmerLookup):
    def count_kmers(self, kmers):
        indices = self._indexed_lookup.find_matches(kmers)
        return np.bincount(indices, minlength=self._kmers.size)

class IndexedSortedLookup(SimpleKmerLookup):
    def __init__(self, sorted_values, index, mod, row_sizes):
        self._sorted_values = np.append(sorted_values, np.array([4**31], dtype=np.uint64))
        self._mod = mod
        self._index = index
        self._row_sizes = row_sizes
        # THis is maybe to memory demanding
        # self._boundries = np.vstack((self._index, self._index+self._row_sizes-1))

    @classmethod
    def from_values_and_mod(cls, values, mod, max_value, do_sort=False):
        if do_sort:
            values = np.sort(values)
        m = values//mod
        index = np.searchsorted(m, np.arange(max_value//mod+2))
        row_sizes = index[1:]-index[:-1]
        return cls(values, index[:-1], mod, row_sizes)

    def find_matches(self, queries):
        row_numbers = queries // self._mod
        row_sizes = self._row_sizes[row_numbers]
        mask = row_sizes > 0
        row_sizes = row_sizes[mask]
        L = self._index[row_numbers[mask]]
        R = L + row_sizes-1
        # boundries = self._boundries[row_numbers[mask]].copy()
        n_iter = int(np.log2(np.max(row_sizes)))+1
        queries = queries[mask]
        found_indices = self._binary_search(queries, L, R, n_iter)
        return found_indices[self._sorted_values[found_indices]==queries]

    def _binary_search(self, queries, L, R, n_iter):
        """
        https://en.wikipedia.org/wiki/Binary_search_algorithm#Alternative_procedure
        """
        boundries = np.vstack((L, R))
        indexes = np.arange(boundries.shape[-1])
        for _ in range(n_iter):
            m = (boundries.sum(axis=0)+1)//2
            is_larger = (self._sorted_values[m] > queries).astype(int)
            boundries[is_larger, indexes]= m-is_larger
            """
            m = (L+R+1)//2
            is_larger = self._sorted_values[m] > queries
            R[is_larger] = m[is_larger]-1
            L[~is_larger] = m[~is_larger]
            """
        return boundries[0] 
                          
    def find_queries(self, queries):
        row_numbers = queries // self._mod
        L = self._index[row_numbers]
        row_sizes = self._row_sizes[row_numbers]
        R = L + row_sizes-1
        m = L
        n_iter = int(np.log2(np.max(row_sizes)))+1
        for _ in range(n_iter):
            m = (L+R+1)//2
            is_larger = self._sorted_values[m] > queries
            R[is_larger] = m[is_larger]-1
            L[~is_larger] = m[~is_larger]
        return L

class HashedKmerLookup(SimpleKmerLookup):
    def count_kmers(self, kmers):
        hashes = kmers % self.mod
        single_hit_mask = self._single_hit_mask[hashes]
        multi_hit_mask =  self._multi_hit_mask[hashes]
        single_hit_idxs = self._hash_to_kmer_idx[single_hit_mask]
        multi_hit_idxs = self._get_indexes(kmers[multi_hit_mask])
        all_idxs = np.concatenate([
            single_hit_idxs[self._kmers[single_hit_idxs]==kmers[single_hit_mask]],
            multi_hit_idxs[self._kmers[multi_hit_idxs]==kmers[multi_hit_mask]]
            ])
        return np.bincount(all_idxs, minlength=self._kmers.size)


        hash_hit_mask = self._hash_to_kmer[hashes] == kmers
        kmers = kmers[n_hits==1] == self._hash_to_kmer[hashes[n_hist==1]]

    def _create_hash_table(self, sorted_kmers):
        hashes = kmers % self.mod
        n_hits = np.bincount(hashes, minlength=mod)
        self._single_hit_mask = n_hits==1
        self._multi_hit_mask = n_hits>1
        self._hash_to_kmer_idx = np.empty_like(n_hits)
        self._hash_to_kmer_idx[hashes] = np.arange(kmers.size)



import time
import logging
from shared_memory_wrapper import SingleSharedArray, to_shared_memory, from_shared_memory
import numpy as np
HEADER = 62
NEWLINE = 10
letters = ["A", "C", "T", "G"]

to_text = lambda x: "".join(chr(r) for r in x)

def get_mask_from_intervals(intervals, size):
    """ intervals = (starts, ends) """
    mask_changes = np.zeros(size+1, dtype="bool")
    mask_changes[intervals[0]]=True
    mask_changes[intervals[1]]^=True
    mask = np.logical_xor.accumulate(mask_changes)
    return mask[:-1]


class Sequences:
    def __init__(self, sequences, intervals_start, intervals_end):
        self.sequences = sequences
        self.offsets = intervals_start[1:]
        self.intervals_start = intervals_start
        self.intervals_end = intervals_end

    def __len__(self):
        return len(self.offsets)+1

    def __getitem__(self, i):
        start, end = (0, len(self.sequences))
        if i > 0:
            start = self.offsets[i-1]
        if i < len(self.offsets):
            end = self.offsets[i]
        return self.sequences[start:end]

    def __repr__(self):
        return "Seqs(%s, %s)" % (to_text(self.sequences), self.offsets)


class TextParser:
    def __init__(self, filename, chunk_size=100):
        self._file_obj = open(filename, "rb")
        self._chunk_size = chunk_size
        self._is_finished = False
        self._last_header_idx = -1

    def read_raw_chunk(self):
        array = np.empty(self._chunk_size, dtype="uint8")
        bytes_read = self._file_obj.readinto(array)
        return array, bytes_read

    def parse(self, as_shared_memory_object=False):
        while not self._is_finished:
            t = time.perf_counter()
            chunk = self.parse_chunk()
            logging.info("Time to parse chunk: %.4f" % (time.perf_counter()-t))
            if chunk is not None:
                if as_shared_memory_object:
                    shared_memory_name = str(np.random.randint(0, 10e15))
                    to_shared_memory(chunk, shared_memory_name)
                    yield shared_memory_name
                else:
                    yield chunk


    def parse_chunk(self):
        a, bytes_read = self.read_raw_chunk()
        self._is_finished = bytes_read < self._chunk_size
        if bytes_read == 0:
            return None
        sequences = self.get_sequences(a[:bytes_read])
        if not self._is_finished:
            self._file_obj.seek(self._last_header_idx-self._chunk_size, 1)
        return sequences

class OneLineFastaParser(TextParser):
    HEADER = 62
    n_lines_per_entry = 2
    def _cut_array(self, array, last_newline):
        self._last_header_idx = last_newline+1
        return array[:last_newline]

    def get_sequences(self, array):
        # print("<--", to_text(array), "-->")
        # print(array.size, array[-1], self._is_finished)
        assert array[0]==self.HEADER
        new_lines = np.flatnonzero(array==NEWLINE)
        if self._is_finished and array[-1] != NEWLINE:
            #Add a trailing newline if this is the last chunk
            new_lines = np.append(new_lines, array.size) 

        # Find which lines are sequences, and cut the array so that it ends with a complete sequence
        is_sequence_line = array[new_lines[:-1]+1] != self.HEADER
        idx_last_sequence_line = np.flatnonzero(is_sequence_line)[-1]
        new_lines = new_lines[:idx_last_sequence_line+2]
        is_sequence_line = is_sequence_line[:idx_last_sequence_line+1]
        array = self._cut_array(array, new_lines[-1])
        sequence_starts = new_lines[:-1][is_sequence_line]+1
        sequence_ends = new_lines[1:][is_sequence_line]
        return self._mask_and_move_sequences(array, sequence_starts, sequence_ends)

    def _mask_and_move_sequences(self, array, sequence_starts, sequence_ends):
        """
        Create a mask for where the sequences are and move sequences to continous array
        """
        mask = get_mask_from_intervals((sequence_starts, sequence_ends), array.size)
        removed_areas = np.cumsum(sequence_starts-np.insert(sequence_ends[:-1], 0, 0))
        new_intervals = (sequence_starts-removed_areas, sequence_ends-removed_areas)
        return Sequences(array[mask], new_intervals[0], new_intervals[1])

"""
class OneLineFastaParser2bit(OneLineFastaParser2bit):
    def _mask_and_move_sequences(self, array, sequence_starts, sequence_ends):
        #Create a mask for where the sequences are and move sequences to continous array
        start_idxs = sequence_starts // 4
        end_idxs = sequence_ends // 4
        mask = get_mask_from_intervals((sequence_starts, sequence_ends), array.size)
        removed_areas = np.cumsum(sequence_starts-np.insert(sequence_ends[:-1], 0, 0))
        new_intervals = (sequence_starts-removed_areas, sequence_ends-removed_areas)
        return Sequences(array[mask], new_intervals[0], new_intervals[1])
"""


class FastaParser(TextParser):
    def _get_sequence_offsets(self, new_lines, is_header):
        header_starts = new_lines[:-1][is_header[:-1]]
        header_ends = new_lines[1:][is_header[:-1]]
        header_lens = header_ends-header_starts
        masked_count = np.ones(new_lines.size, dtype="int")
        masked_count[1:][is_header[:-1]] = header_lens
        masked_count[0] = new_lines[0]+1
        cum_masked_count = np.cumsum(masked_count)
        indexes_after_masking = new_lines+1-cum_masked_count
        starts = indexes_after_masking[1:][is_header[:-1]]
        if not self._is_finished:
            starts = starts[:-1]
        return starts

    def get_sequences(self, array):
        assert array[0] == HEADER
        all_new_lines = np.flatnonzero(array==NEWLINE)
        new_lines = all_new_lines if all_new_lines[-1] != array.size-1 else all_new_lines[:-1]
        is_header = array[new_lines+1] == HEADER
        mask_changes = np.zeros(array.size, dtype="bool")
        mask_changes[new_lines+1] = np.where(is_header, 1, 0)
        mask_changes[new_lines[1:]] = np.where(is_header[:-1], 1, 0)
        mask_changes[0] = 1
        mask_changes[new_lines[0]] = 1
        mask = np.logical_xor.accumulate(mask_changes)
        mask[all_new_lines] = 1
        beginnings = self._get_sequence_offsets(new_lines, is_header)
        intervals = (np.insert(beginnings, 0, 0), np.append(beginnings, np.count_nonzero(~mask)))
        if self._is_finished:
            return Sequences(array[~mask], intervals)
        assert np.any(is_header), ("Fasta entry to big for chunk size", to_text(array), chunk_size)
        self._last_header_idx = new_lines[is_header][-1]+1
        intervals = (np.insert(beginnings, 0, 0), np.append(beginnings, np.count_nonzero(~mask[:self._last_header_idx])))
        return Sequences(array[:self._last_header_idx][~mask[:self._last_header_idx]], intervals)


class KmerHash:
    CODES = np.zeros(256, dtype=np.uint64)
    CODES[[97, 65]] = 0
    CODES[[99, 67]] = 1
    CODES[[116, 84]] = 2
    CODES[[103, 71]] = 3

    def __init__(self, k=31):
        self.k=k
        self.POWER_ARRAY = np.power(4, np.arange(k, dtype=np.uint64), dtype=np.uint64)
        self.REV_POWER_ARRAY = np.power(4, np.arange(k, dtype=np.uint64)[::-1], dtype=np.uint64)


    def _to_text(self, seq):
        return "".join(letters[n] for n in seq)

    def get_mask(self, offsets, size):
        """
        k=3
        offsets=[5, 9]
        0123456789
        ACGTTGATTCGGA
        -----s---s---
        ---          0
         ---         1
          ---        2
           --x       3 = 5-k+1
            -x-      4 
             ---     5 = 5 
              ---    6
        """

        mask_diffs = np.zeros(size+1, dtype="int")
        mask_diffs[np.maximum(offsets-self.k+1, 0)] = 1
        mask_diffs[np.minimum(offsets, size)] = -1
        return np.cumsum(mask_diffs)[:-1]==0

    def _get_kmers(self, seq):
        return [self._to_text(seq[i:i+self.k]) for i in range(len(seq)-self.k+1)]

    def _get_codes2(self, seq):
        idxs = np.digitize((seq-65) % 32 , np.array([65, 71, 97, 103]), right=True)
        return np.asarray(idxs^(idxs>>1), dtype=np.uint64)

    def _get_codes(self, seq):
        return self.CODES[seq]

    def _join(self, kmers1, kmers2, mask):
        return np.concatenate((kmers1[mask], kmers2[mask]))

    def get_kmer_hashes(self, sequences, include_reverse_complements=True):
        codes = self._get_codes(sequences.sequences)
        assert codes.dtype==np.uint64, codes.dtype
        kmers = np.convolve(codes, self.POWER_ARRAY, mode="valid")
        mask = self.get_mask(sequences.offsets, kmers.size)
        reverse_kmers = None

        if include_reverse_complements:
            reverse_kmers = np.convolve((codes+2) % 4, self.REV_POWER_ARRAY, mode="valid")

        return kmers, reverse_kmers, mask # [mask], reverse_kmers[mask]

def simulate_fasta(n_seqs, seq_len, filename):
    f = open(filename, "w")
    for i in range(n_seqs):
        seq = "".join(letters[n] for n in np.random.randint(0, 4, seq_len))
        f.write(f">h{i}\n{seq}\n")
    f.close()

def main(filename, chunk_size=500*10**5, do_print=False):
    parser = OneLineFastaParser(filename, chunk_size)
    for i, seqs in enumerate(parser.parse()):
        if do_print:
            for i in range(len(seqs)):
                print(to_text(seqs[i]))
        # print(f"Parsing chunk {i}")
        kmers = KmerHash(3).get_kmer_hashes(seqs)

if __name__ == "__main__":
    import sys
    import cProfile
    #simulate_fasta(10, 20, "small.fa")
    #cProfile.run('main("large.fa")')
    # exit()
    main(sys.argv[1], int(sys.argv[2]), do_print=True)
    exit()
    # for seqs in :# parse_fasta(, 100):
    #     print(seqs)
    #     kmers = KmerHash(3).get_kmer_hashes(seqs)
    #     continue
    #     t, starts = (seqs.sequencegs, seqs.offsets)
    #     print(to_text(t), starts)
    #     print("<", to_text(t[:starts[0]]))
    #     for s,e in zip(starts[:-1], starts[1:]):
    #         print("-", to_text(t[s:e]))
    #     print(">", to_text(t[starts[-1]:]))



import logging
import sys
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s')

from graph_kmer_index.collision_free_kmer_index import MinimalKmerIndex

import numpy as np
import argparse
from graph_kmer_index import KmerIndex
from .mapping import get_kmers_from_fasta, map_fasta
from mapper import map_kmers_to_graph_index
import time
from graph_kmer_index.index_bundle import IndexBundle
from .kmer_counting import SimpleKmerLookup
from .kmer_lookup import Advanced2
from .parser import OneLineFastaParser, KmerHash, Sequences
from shared_memory_wrapper import from_shared_memory, to_shared_memory, SingleSharedArray
from shared_memory_wrapper.shared_memory import remove_shared_memory_in_session
from pathos.multiprocessing import Pool
from itertools import repeat

def main():
    run_argument_parser(sys.argv[1:])

def test(args):
    print("Hi")


def map_fasta_command(args):
    if args.kmer_size > 31:
        logging.error("k must be 31 or lower")
        sys.exit(1)

    if not args.fasta_file.endswith(".fa"):
        logging.error("Only fasta files (not fq or gzipped files) are supported for now.")
        sys.exit(1)


    if args.kmer_index is None:
        if args.index_bundle is None:
            logging.error("Either a kmer index (-i) or an index bundle (-b) needs to be specified")
            sys.exit(1)
        else:
            kmer_index = IndexBundle.from_file(args.index_bundle).indexes
    else:
        if args.use_numpy:
            logging.info("Using numpy index")
            args.kmer_index = Advanced2.from_old_index_files(args.kmer_index)
        else:
            cls = KmerIndex
            if "minimal" in args.kmer_index:
                cls = MinimalKmerIndex
            kmer_index = cls.from_file(args.kmer_index)

    map_fasta(args, kmer_index)


def run_argument_parser(args):

    parser = argparse.ArgumentParser(
        description='Kmer Mapper',
        prog='kmer_mapper',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50, width=100))

    subparsers = parser.add_subparsers()
    subparser = subparsers.add_parser("map")
    subparser.add_argument("-i", "--kmer-index", required=False)
    subparser.add_argument("-b", "--index-bundle", required=False)
    subparser.add_argument("-f", "--fasta-file", required=True)
    subparser.add_argument("-k", "--kmer-size", required=False, default=31, type=int)
    subparser.add_argument("-t", "--n-threads", required=False, default=16, type=int)
    subparser.add_argument("-c", "--chunk-size", required=False, type=int, default=500000, help="N reads to process in each chunk")
    subparser.add_argument("-n", "--use-numpy", required=False, type=bool, default=False, help="Use numpy-based counting instead of Cython")
    subparser.add_argument("-N", "--use-cython-file-reading", required=False, type=bool, default=False, help="Use cython-based file reading instead of Cython")
    subparser.add_argument("-m", "--no-shared-memory", required=False, type=bool, default=False, help="Set to True to not use shared memory for index. Increases memory usage by a factor of --n-threads.")
    subparser.add_argument("-l", "--max-read-length", required=False, type=int, default=150,
                           help="Maximum length of reads. Reads should not be longer than this.")
    subparser.add_argument("-o", "--output-file", required=True)
    subparser.add_argument("-r", "--include-reverse-complement", required=False, default=False, type=bool)
    subparser.add_argument("-I", "--max-hits-per-kmer", required=False, default=1000, type=int,
                           help="Ignore kmers that have more than this amount of hits in index")
    subparser.set_defaults(func=map_fasta_command)


    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args)
    args.func(args)

    remove_shared_memory_in_session()
